import os
import re

from astra_installer.commands import run_command
from astra_installer.logs.logger import Logger
from astra_installer.settings.manager import SettingsManager
from astra_installer.setup.common.disk import DiskInstall
from astra_installer.storage.utils.storage import get_device_uuid


class DiskInstallE2K8C(DiskInstall):
    def __init__(self) -> None:
        super().__init__()

    def hardware_revision(self) -> str:
        Logger.log('Finding hardware revision')
        loaded_kernel = run_command('uname -r').stdout[0]
        match = re.match(r'^.*?-(?P<revision>\d+c)$', loaded_kernel)
        if match and (match.group('revision') == '1c' or match.group('revision') == '8c'):
            Logger.log('Hardware revision {revision}'.format(revision=match.group('revision')))
            return match.group('revision')
        else:
            raise RuntimeError('Unknown hardware revision. Loaded kernel {kernel}'.format(kernel=loaded_kernel))

    def install_kernel(self) -> None:
        Logger.log('Installing kernel and modules')
        version = '4.9.0-2'
        revision = self.hardware_revision()
        kernel = 'linux-image-{version}-generic-{revision}'.format(version=version, revision=revision)
        modules = 'linux-astra-modules-generic-{revision}-meta'.format(revision=revision)
        self.chroot.apt_get_install([kernel, modules])
        self.make_symlinks(version, revision)

    def make_symlinks(self, version: str, revision: str) -> None:
        Logger.log('Making kernel symlinks')
        vmlinuz = 'image-{version}-generic-{revision}'.format(version=version, revision=revision)
        initrd = 'initrd.img-{version}-generic-{revision}'.format(version=version, revision=revision)
        boot_path = os.path.join(SettingsManager.disk.target_path, 'boot')
        os.symlink(vmlinuz, os.path.join(boot_path, 'vmlinuz'))
        os.symlink(initrd, os.path.join(boot_path, 'initrd.img'))

    def write_bootconf(self) -> None:
        Logger.log('Writing boot.conf')
        text = """default=Astra
timeout=3

########################

label=Astra
        partition=0
        image=/vmlinuz
        initrd=/initrd.img
        cmdline=console=tty0 consoleblank=0 hardreset root=UUID={root_uuid} net.ifnames=0 ipv6.disable=1 parsec.max_ilev=0

label=Recovery
        partition=0
        image=/vmlinuz
        initrd=/initrd.img
        cmdline=console=tty0 consoleblank=0 hardreset root=UUID={root_uuid} net.ifnames=0 ipv6.disable=1 parsec.max_ilev=0 init=/bin/bash
""".format(root_uuid=get_device_uuid(SettingsManager.storage.current_template.storage_state.root_partition().path))
        self.chroot.create_file('/boot/boot.conf', text)

    def setup_bootloader(self) -> None:
        self.install_kernel()
        self.write_bootconf()
