from typing import List, Optional, Dict, TYPE_CHECKING

import parted
from blivet.actionlist import ActionList
from blivet.devices import DiskDevice
from blivet.devices import StorageDevice
from blivet.partitioning import PartitionDevice
from blivet.size import Size
from pyudev import Device

from astra_installer_kickstart.module.partitioning_module import PartitioningModule
from astra_installer.storage.utils.storage import mounted_devices

if TYPE_CHECKING:
    from gettext import gettext as _

min_partition_sizes = {'/': Size('250MB'),
                       '/usr': Size('250MB'),
                       '/tmp': Size('50MB'),
                       '/var': Size('384MB'),
                       '/home': Size('100MB'),
                       '/boot': Size('250MB')}

min_swap_partition_size = Size('256MB')


class MountData:
    def __init__(self, target: str, mountpoint: str):
        self.__target = target
        self.__mountpoint = mountpoint

    @property
    def target(self) -> str:
        return self.__target

    @property
    def mountpoint(self) -> str:
        return self.__mountpoint


class StorageState:
    def __init__(self, partition_module: PartitioningModule, based_on_device: Device = None) -> None:
        self.__partition_module = partition_module
        self.__based_on_device = based_on_device
        self.__based_on_gpt = False
        self.__bios_is_used = False

    @property
    def actions(self) -> ActionList:
        return self.__partition_module.storage.devicetree.actions

    @property
    def partition_module(self) -> PartitioningModule:
        return self.__partition_module

    @property
    def based_on_device(self) -> Optional[Device]:
        return self.__based_on_device

    @based_on_device.setter
    def based_on_device(self, device: Optional[Device]) -> None:
        self.__based_on_device = device

    @property
    def based_on_gpt(self) -> bool:
        return self.__based_on_gpt

    @based_on_gpt.setter
    def based_on_gpt(self, used: bool) -> None:
        self.__based_on_gpt = used

    @property
    def bios_is_used(self) -> bool:
        return self.__bios_is_used

    @bios_is_used.setter
    def bios_is_used(self, used: bool) -> None:
        self.__bios_is_used = used

    @property
    def mountpoints(self) -> Dict[str, PartitionDevice]:
        return self.__partition_module.storage.fsset.mountpoints

    def apply(self) -> None:
        self.__partition_module.storage.devicetree.teardown_all()
        self.__partition_module.storage.do_it()
        self.__setup_bootable_devices(self.__partition_module.storage)
        self.__partition_module.storage.dump_state("final")
        # self.__partition_module.storage.turn_on_swap()

    def mounting_partitions(self) -> List[MountData]:
        mount_list = []  # type: List[MountData]

        for mountpoint, device in self.mountpoints.items():
            mount_list.append(MountData(target=device.path, mountpoint=mountpoint))

        return mount_list

    def efi_partition(self) -> Optional[DiskDevice]:
        return self.__partition_module.storage.efi_partition

    def bootable_device(self) -> Optional[DiskDevice]:
        return self.__partition_module.storage.bootloader.stage1_disk or self.__partition_module.storage.boot_device

    def root_partition(self) -> Optional[PartitionDevice]:
        return self.__partition_module.storage.root_device

    def fstab_text(self) -> str:
        return self.__partition_module.storage.fsset.fstab()

    def write_configs(self) -> str:
        return self.__partition_module.storage.fsset.write()

    def used_devices(self) -> List[StorageDevice]:
        devices = []
        for action in self.__partition_module.storage.devicetree.actions:
            if action.device.exists and (action.is_destroy or action.is_format or action.is_remove or action.is_create):
                devices.append(action.device)
                devices += action.device.children

        return devices

    def cleared_partitions(self) -> List[PartitionDevice]:
        destroy_actions = self.__partition_module.storage.devicetree.actions.find(
            action_type="destroy",
            object_type="device"
        )

        partitions = [
            action.device for action in destroy_actions
            if action.device.type == "partition"
        ]

        return partitions

    def validate(self, validate_efi: bool = True) -> Optional[str]:
        from astra_installer.settings.manager import SettingsManager

        used_devices = set([device.path for device in self.used_devices()])
        used_devices |= set([device.path for mountpoint, device in self.mountpoints.items()
                             if device.path not in used_devices])

        conflicted_devices = mounted_devices() & used_devices
        if conflicted_devices:
            return _('Partitions {partitions} are mounted. Please unmount partitions') \
                .format(partitions=', '.join(conflicted_devices))

        if not self.root_partition():
            return _('You must specify at least one partition with "{mountpoint}" mountpoint') \
                .format(mountpoint='/')

        if not self.bootable_device():
            return _('You must specify one disk as bootable')

        if SettingsManager.systeminfo.efi and validate_efi and not self.bios_is_used:
            if not self.efi_partition():
                return _('You must specify one EFI system partition with "{mountpoint}" mountpoint') \
                    .format(mountpoint='/boot/efi')

        return None

    def warnings(self) -> List[str]:
        warnings = []

        for mountpoint, device in self.mountpoints.items():
            min_size = min_partition_sizes.get(mountpoint)
            if min_size is None:
                continue

            if device.size < min_size:
                warnings.append(_('Partition with "{mountpoint}" mountpoint with {size} size is smaller than '
                                  'recommended {min_size} size').format(mountpoint=mountpoint,
                                                                        size=device.size.human_readable(),
                                                                        min_size=min_size.human_readable()))

        for device in self.__partition_module.storage.fsset.fstab_swaps:
            if device.size < min_swap_partition_size:
                warnings.append(_('Swap partition with {size} size is smaller than recommended '
                                  '{min_size} size').format(size=device.size.human_readable(),
                                                            min_size=min_swap_partition_size.human_readable()))

        return warnings

    def __setup_bootable_devices(self, storage) -> None:
        """Set up the bootable devices.

        Mark the boot devices as bootable.

        :param storage: an instance of the storage
        """
        if storage.bootloader.skip_bootloader:
            return

        if storage.bootloader.stage2_bootable:
            boot = storage.boot_partition
        elif storage.bootloader.stage1_device:
            boot = storage.bootloader.stage1_device
        else:
            boot = storage.boot_partition

        # if storage.bootloader.stage2_bootable:
        #     boot = storage.boot_device
        # else:
        #     boot = storage.bootloader.stage1_device

        if boot.type == "mdarray":
            boot_devs = boot.parents
        else:
            boot_devs = [boot]

        for dev in boot_devs:
            if not hasattr(dev, "bootable"):
                # log.info("Skipping %s, not bootable", dev)
                continue

            # Dos labels can only have one partition marked as active
            # and unmarking ie the windows partition is not a good idea
            skip = False
            if dev.disk.format.parted_disk.type == "msdos":
                for p in dev.disk.format.parted_disk.partitions:
                    if p.type == parted.PARTITION_NORMAL and \
                            p.getFlag(parted.PARTITION_BOOT):
                        skip = True
                        break

            # GPT labeled disks should only have bootable set on the
            # EFI system partition (parted sets the EFI System GUID on
            # GPT partitions with the boot flag)
            if dev.disk.format.label_type == "gpt" and \
                    dev.format.type not in ["efi", "macefi"]:
                skip = True

            if skip:
                # log.info("Skipping %s", dev.name)
                continue

            # hfs+ partitions on gpt can't be marked bootable via parted
            if dev.disk.format.parted_disk.type != "gpt" or \
                    dev.format.type not in ["hfs+", "macefi"]:
                # log.info("setting boot flag on %s", dev.name)
                dev.bootable = True

            # Set the boot partition's name on disk labels that support it
            if dev.parted_partition.disk.supportsFeature(parted.DISK_TYPE_PARTITION_NAME):
                ped_partition = dev.parted_partition.getPedPartition()
                ped_partition.set_name(dev.format.name)
                # log.info("Setting label on %s to '%s'", dev, dev.format.name)

            dev.disk.setup()
            dev.disk.format.commit_to_disk()
